/**
* @file   Array.h
* @author Morten Hjorth-Jensen (morten.hjorth-jensen@fys.uio.no) and Islen Vallejo
(islenv@fys.uio.no)
* @date   May, 2010
* @class  Array
* @brief  Class used for manipulating up to two-dimensional arrays.
*
* Contains user-defined operators to do computations with arrays in a style 
* close to mathematical equations. 
* 
* \remarks Because this is template class, the source file @ref Array.cpp must 
* be included in * the header file in order to avoid problems during linking stage. 
* It is what we have done in line 249. 
*
* \remarks We do not need compile the source file @ref Array.cpp because it has
* already been include in the header file.
* 
*
* Example Usage:
@code


#include "Array.h"

#include <iostream>
using namespace std;

int main(){

  // Create a array with (default) nrows = 1, ncols = 1:
  Array<double> v1;

  // Redimension the array to have length n:
  int n1 = 3;
  v1.redim(n1);

  // Extract the length of the array:
  const int length = v1.getLength();

  // Create a array of a specific length:
  int n2 = 5;
  Array<double> v2(n2);

  // Create a array as a copy of another one:
  Array<double> v5(v1);

  // Assign the entries in a array:
  v5(0) = 3.0;  
  v5(1) = 2.5;  
  v5(2) = 1.0;  

  for(int i=0; i<3; i++){
	  cout << v5(i) << endl;
  }
  
  // Extract the ith component of a array:
  int i = 2;
  double value = v5(1);
  cout << "value: " << value << endl;

  // Set a array equal another one:
  Array<double> v6 = v5;

  for(int i=0; i<3; i++){
	  v1(i) = 1.0;
	  v2(i) = 2.0;
  }
  
  // Create a two-dimensional array (matrix):
  Array<double> matrix(2, 2);
  
  // Fill the array:
  matrix(0,0) = 1;
  matrix(0,1) = 2;
  matrix(1,0) = 3;
  matrix(1,1) = 4;
  
  // Get the entries in the array:
  cout << "\nMatrix: " << endl;
  for(int i=0; i<2; i++){
    for(int j=0; j<2; j++){
      cout << matrix(i,j) << "   ";
    }
    cout << endl;
  }
  
  // Assign a entry of the matrix to a variable:
  double scalar = matrix(0,0);
  const double b = matrix(1,1);
   
    
  Array<double> vector(2);
  vector(0) = 1.0; 
  vector(1) = 2.0;
    
  Array<double> v = vector;
  Array<double> A = matrix;
  Array<double> u(2);
  
  cout << "\nMatrix: " << endl;
  for(int i=0; i<2; i++){
    for(int j=0; j<2; j++){
      cout << matrix(i,j) << "   ";
    }
    cout << endl;
  }
  
  Array<double> a(2,2);
  a(1,1) = 5.0;
  
  // Arithmetic operations with arrays using a 
  // syntax close to the mathematical language
  Array<double> w = v1 + 2.0*v2;
  
	// Create multidimensional matrices and assign values to them:
  int N = 3;
  Array<double> multiD;  multiD.redim(N,N,N);
  for(int i=0; i<N; i++){
	  for(int j=0; j<N; j++){
		  for(int k=0; k<N; k++){
			  cout << "multD(i,j,k) = " << multiD(i,j,k) << endl;
		  }
	  }
  }

  multiD(1,2,3) = 4.0;
  cout << "multiD(1,2,3) = " << multiD(1,2,3) << endl;

  // Array<double> can also send data to Python/Fortran/C functions. For example, if v
  // is a array object, what we do is to send an array by doing: v.getPtr();  
}
@endcode
*
*
**/




#ifndef ARRAY_H
#define ARRAY_H

#include <iostream>
#include <sstream>
#include <iomanip>
#include <cstdlib>

using namespace std;

template<class T>
class Array{
  private:
    static const int MAXDIM = 6;
		T *data ;    				/**> One-dimensional array of data.*/
  	int size[MAXDIM];		/**> Size of each dimension.*/
		int ndim;						/**> Number of dimensions occupied. */
		int length;					/**> Total number of entries.*/
  
		int dx1, dx2, dx3, dx4, dx5;	
  
		void allocate(int ni=0, int nj=0, int nk=0, int nl=0, int nm=0, int nn=0){
			ndim = MAXDIM;
			
			// Set the number of entries in each dimension.
			size[0]=ni;
			size[1]=nj;
			size[2]=nk;
			size[3]=nl;
			size[4]=nm;
			size[5]=nn;			
			
			
			// Set the number of dimensions used.
			if(size[5] == 0)
				ndim--;
			if(size[4] == 0)
				ndim--;
			if(size[3] == 0)
				ndim--;
			if(size[2] == 0)
				ndim--;
			if(size[1] == 0) 
				ndim--;
			if(size[0] == 0){
				ndim 		= 0;
				length 	= 0;
				data   	= NULL;
			}else{
				try{
					int i;
							
					// Set the length (total number of entries) of the one-dimensional array.
					length = 1;
					for(i=0; i<ndim; i++)
						length *= size[i];
						
						data = new T[length];
					
						dx1 = 		size[0];
						dx2 = dx1*size[1];
						dx3 = dx2*size[2];
						dx4 = dx3*size[3];
						dx5 = dx4*size[4];
											
				}catch(std::bad_alloc&){
					std::cerr << "Array::allocate -- unable to allocate array of length " << length << std::endl;
					exit(1);
				}
			}
			
		}
    
  public:
          
		
    /**
		* @brief Constructor with default arguments. 
		*
		* Creates an array with one or two-dimensions.
		*
		* @param int nrows. Number of rows in the array.
		* @param int ncolsd. Number of columns in the array.
		**/
    Array(int ni=0, int nj=0, int nk=0, int nl=0, int nm=0, int nn=0){
			// Allocate memory
			allocate(ni,nj,nk,nl,nm,nn);
		} // end constructor
			
			
		//! Constructor	
		Array(T* array, int ndim_, int size_[]){
			ndim = ndim_;
			
		  length = 1;
			int i;
			for(i=0; i<ndim; i++){
				size[i] = size_[i];	// Copy only the ndim entries. The rest is zero by default.
				length *= size[i];
			}
							
			// Now when we known the length, we should not forget to allocate memory!!!!
			data = new T[length];
			
			// Copy the entries from array to data:
			for(i=0; i<length; i++){
			  data[i] = array[i];
			}
			
		} // End constructor.
		
		
			
		
		//! Copy constructor
    Array(const Array<T>& array);    
    
    //! Destructor
    ~Array();
				
		
		/**
		* @brief Checks the validity of the indexing.
		* @param i, an integer for indexing the rows.
		* @param j, an integer for indexing the columns.
		**/
		bool indexOk(int i, int j=0) const;    
    
    /**
		* @brief Change the dimensions of an array.
		* @param ni number of entries in the first dimension.
		* @param nj number of entries in the second dimension.
		* @param nk number of entries in the third dimension.
		*	@param nl number of entries in the fourth dimension.
		* @param nm number of entries in the fifth dimension.
		* @param nn number of entries in the sixth dimension.
		**/
    bool redim(int ni, int nj=0, int nk=0, int nl=0, int nm=0, int nn=0);		
		
		/**
		* @return The total number of entries in the array, i.e., the sum of the entries in all the dimensions.
		**/
		int getLength()const{return length;}
		
		
		/**
		* @return The number of rows in a matrix.
		**/
		int getRows() const {return size[0];}
		
		/**
		* @return Returns the number of columns in a matrix.
		**/
    int getColumns() const {return size[1];}
    
    
    /** @brief Gives the number of entries in a dimension.
		*
		*	@param i An integer from 0 to 5 indicating the dimension we want to explore. 
		*	@return size[i] An integer for the number of elements in the dimension number i.
		**/		
		int dimension(int i) const{return size[i];}
		
		
		/**
		* The number of dimensions in the array.
		**/
		int getNDIM()const{return ndim;} 
    
    
    /** 
		* @return A constant pointer to the array of data.
		* This function can be used to interface C++ with Python/Fortran/C.
		**/
		const T* getPtr() const;
    
		
		/**
		* @return A pointer to the array of data.
		* This function can be used to interface C++ with Python/Fortran/C.
		**/
		T* getPtr();
		
		
		/**
		*	@return A pointer to an array with information on the length of each dimension. 
		**/
		int* getPtrSize();
		
		
    	
    
		/************************************************************/
		/*											OPERATORS														*/
		/************************************************************/
		
    //! Assignment operator
    Array<T>& operator=(const Array<T>& array);

    //! Sum operator
    Array<T> operator+(const Array<T>& array);
      
    //! Substraction operator    
		Array<T> operator-(const Array<T>& array)const; /// w=u-v;
		
		
		//! Multiplication operator
    //Array<T> operator*(const Array<T>& array);

		
		//! Assigment by addition operator
		Array<T>& operator+=(const Array<T>& w);


		//! Assignment by substraction operator
    Array<T>& operator-=(const Array<T>& w);
    
    
    //! Assignment by scalar product operator
    Array<T>& operator*=(double scalar);
		
    //! Assignment by division operator
		Array<T>& operator/=(double scalar);	
		
    //! Index operators
    const T& operator()(int i)const;	
    const T& operator()(int i, int j)const;	
		const T& operator()(int i, int j, int k)const;	
		const T& operator()(int i, int j, int k, int l)const;		
		const T& operator()(int i, int j, int k, int l, int m)const;		
		const T& operator()(int i, int j, int k, int l, int m, int n)const;		
		
    T& operator()(int i);
		T& operator()(int i, int j);
		T& operator()(int i, int j, int k);	
		T& operator()(int i, int j, int k, int l);		
		T& operator()(int i, int j, int k, int l, int m);		
		T& operator()(int i, int j, int k, int l, int m, int n);		
		
		
   		
		/**************************************************************/
		/*								FRIEND FUNCTIONS 														*/
		/**************************************************************/
 		//! Unary operator +
		template <class T2>
    friend Array<T> operator+ (const Array<T>&);                 // u = + v
 
		//! Unary operator -
		template <class T2>
    friend Array<T> operator-(const Array<T>&);                 // u = - v
    
		
			
		
		/**
    * Premultiplication by a floating point number: 
    * \f$\mathbf{u} = a \mathbf{v}\f$, 
    * where \f$a\f$ is a scalar and \f$\mathbf{v}\f$ is a array.
    **/
		template <class T2>
    friend Array<T> operator*(double, const Array<T>&);         // u = a*v
    
    /**
    * Postmultiplication by a floating point number: 
    * \f$\mathbf{u} = \mathbf{v} a\f$, 
    * where \f$a\f$ is a scalar and \f$\mathbf{v}\f$ is a array.
    **/
		template <class T2>
    friend Array<T> operator*(const Array<T>&, double);         // u = v*a
  
		
		/**
    * Division of the entries of a array by a scalar.
    **/
		template <class T2>
    friend Array<T> operator/(const Array<T>&, double);         // u = v/a 
		
		
		
		
		
};

#include "Array.cpp"


// Destructor
template <class T>
inline Array<T>::~Array(){delete[] data;}

// Index operators
template <class T>
inline const T& Array<T>::operator()(int i)const {
	#if CHECKBOUNDS_ON
		indexOk(i);
	#endif
	return data[i];
}



template <class T>
inline const T& Array<T>::operator()(int i, int j)const {
	#if CHECKBOUNDS_ON
		indexOk(i,j);
	#endif
	
	return data[i + j*dx1];
}

template <class T>
inline const T& Array<T>::operator()(int i, int j, int k)const {
	#if CHECKBOUNDS_ON
		indexOk(i,j,k);
	#endif
	
	return data[i + j*dx1 + k*dx2];
}


template <class T>
inline const T& Array<T>::operator()(int i, int j, int k, int l)const {
	#if CHECKBOUNDS_ON
		indexOk(i,j,k,l);
	#endif
	
	return data[i + j*dx1 + k*dx2 + l*dx3];
}


template <class T>
inline const T& Array<T>::operator()(int i, int j, int k, int l, int m)const {
	#if CHECKBOUNDS_ON
		indexOk(i,j,k,l, m);
	#endif
	
	return data[i + j*dx1 + k*dx2 + l*dx3 + m*dx4];
}


template <class T>
inline const T& Array<T>::operator()(int i, int j, int k, int l, int m, int n)const {
	#if CHECKBOUNDS_ON
		indexOk(i,j,k,l,m,n);
	#endif
	
	return data[i + j*dx1 + k*dx2 + l*dx3 + m*dx4 + n*dx5];
}
	
	
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

	
template <class T>
inline T& Array<T>::operator()(int i) {
	#if CHECKBOUNDS_ON
		indexOk(i);
	#endif
	return data[i];
}
	

template <class T>
inline T& Array<T>::operator()(int i, int j) {
	#if CHECKBOUNDS_ON
		indexOk(i,j);
	#endif
	
	return data[i + j*dx1];
}


template <class T>
inline T& Array<T>::operator()(int i, int j, int k) {
	#if CHECKBOUNDS_ON
		indexOk(i,j,k);
	#endif
	
	return data[i + j*dx1 + k*dx2];
}


template <class T>
inline T& Array<T>::operator()(int i, int j, int k, int l) {
	#if CHECKBOUNDS_ON
		indexOk(i,j,k,l);
	#endif
	
	return data[i + j*dx1 + k*dx2 + l*dx3];
}


template <class T>
inline T& Array<T>::operator()(int i, int j, int k, int l, int m) {
	#if CHECKBOUNDS_ON
		indexOk(i,j,k,l,m);
	#endif
	
	return data[i + j*dx1 + k*dx2 + l*dx3 + m*dx4];
}


template <class T>
inline T& Array<T>::operator()(int i, int j, int k, int l, int m, int n) {
	#if CHECKBOUNDS_ON
		indexOk(i,j,k,l,m,n);
	#endif
	
	return data[i + j*dx1 + k*dx2 + l*dx3 + m*dx4 + n*dx5];
}



template <class T>
inline const T* Array<T>::getPtr() const {return data;}



template <class T>
inline T* Array<T>::getPtr(){return data; }


template <class T>
inline int* Array<T>::getPtrSize(){return size;}


// template <class T>
// inline int Array<T>::dim()const{return ndim;}


/******************************************************************/
/*							IMPLEMENTATION OF FRIEND FUNCTIONS								*/
/******************************************************************/

/******************************************************************/
/*             (Arithmetic) Unary operators                       */
/******************************************************************/
//! Unary operator +
template <class T>
inline Array<T> operator+(const Array<T>& v){     // u = + v
	return v;
}


//! Unary operator -
template <class T>
inline Array<T> operator-(const Array<T>& v){      // u = - v
	return Array<T>(v.size[0],v.size[1]) -v;
}


//! Postmultiplication operator
template <class T>
inline Array<T> operator*(const Array<T>& v, double scalar){   // u = v*a
  return Array<T>(v) *= scalar;
}


//! Premultiplication operator. 
template <class T>
inline Array<T> operator*(double scalar, const Array<T>& v){   // u = a*v
  return v*scalar;  // Note the call to postmultiplication operator defined above
}


//! Division of the entries in a array by a scalar
template <class T>
inline Array<T> operator/(const Array<T>& v, double scalar){ 
  if(!scalar) std::cout << "Division by zero!" << std::endl;
  return (1.0/scalar)*v;
}

#endif 
